import sys
from notario.exceptions import Invalid, SchemaError
from notario.utils import is_callable


class Validator(object):

    def __init__(self, data, schema):
        self.data = normalize(data)
        self.schema = normalize_schema(schema)

    def validate(self):
        self.length_equality(self.data, self.schema, 0, [])
        self.traverser(self.data, self.schema, [])

    def traverser(self, data, schema, tree):
        """
        Traverses the dictionary, recursing onto itself if
        it sees apropriate key/value pairs that indicate that
        there is a need for more validation in a branch below us.
        """
        if hasattr(schema, '__validator_leaf__'):
            return schema(data, tree)

        for index in range(len(data)):
            self.length_equality(data, schema, index, tree)
            key, value = data[index]
            skey, svalue = schema[index]
            tree.append(key)
            if isinstance(value, dict):
                self.traverser(value, svalue, tree)
            else:
                self.leaf(data[index], schema[index], tree)
            if tree:
                tree.pop()

    def leaf(self, data, schema, tree):
        """
        The deepest validation we can make in any given circumstance. Does not
        recurse, it will just receive both values and the tree, passing them on
        to the :fun:`enforce` function.
        """
        key, value = data
        skey, svalue = schema
        enforce(key, skey, tree, 'key')
        if hasattr(svalue, '__validator_leaf__'):
            return svalue(value, tree)
        enforce(value, svalue, tree, 'value')

    def length_equality(self, data, schema, index, tree):
        try:
            data = data[index]
            schema = schema[index]
        except (KeyError, TypeError):
            raise SchemaError(data, tree, reason="has less items in schema than in data")
        if hasattr(schema, '__validator_leaf__'):
            return
        if len(data) != len(schema):
            raise SchemaError(data, tree, reason='length did not match schema')


class BaseItemValidator(object):

    def __init__(self, data, schema, tree=None, index=None):
        self.data = data
        self.schema = schema
        self.tree = tree or []
        self.index = index or 0

    def validate(self):
        self.traverser(self.data, self.schema, self.tree)

    def traverser(self, data, schema, tree):
        if len(data) < self.index:
            raise SchemaError(data, tree, reason="has not enough items to select from")
        self.leaves(data, schema, tree)


class IterableValidator(BaseItemValidator):
    """
    The iterable validator allows the definition of a single schema that can be
    run against any number of items in a given data structure
    """

    def data_sanity(self, data):
        if not isinstance(data, list):
            raise SchemaError('', [], reason='IterableValidator needs a list to validate', pair='value')

    def leaf(self, index):
        self.data_sanity(self.data)
        self.enforce(self.data, self.schema, index, self.tree)

    def leaves(self, data, schema, tree):
        self.data_sanity(data)
        for item_index in range(self.index, len(data)):
            self.enforce(data, schema, item_index, tree)

    def enforce(self, data, schema, item_index, tree):
        if isinstance(data[item_index], dict) and isinstance(schema, tuple):
            try:
                _validator = Validator(data[item_index], schema)
                _validator.validate()
            except Invalid:
                e = sys.exc_info()[1]
                tree.append('list[%s]' % item_index)
                tree.extend(e.path)
                raise Invalid(e.schema_item, tree, pair='value')
            except SchemaError: # FIXME this is utterly redundant, and also happens in RecursiveValidator
                e = sys.exc_info()[1]
                tree.extend(e.path)
                raise SchemaError('', tree, reason=e._reason, pair='value')

        elif isinstance(schema, tuple) and not isinstance(data[item_index], (tuple, dict)):
            raise SchemaError(data, tree, reason='iterable contains single items, schema does not')
        else:
            try:
                if is_callable(schema):
                    schema(data[item_index])
                else:
                    assert data[item_index] == schema
            except AssertionError:
                tree.append('list[%s]' % item_index)
                raise Invalid(schema, tree, pair='item')


class RecursiveValidator(BaseItemValidator):
    """
    The recursive validator allows the definition of a single schema that can be
    run against any number of items in a given data structure
    """

    def leaf(self, index):
        self.enforce(self.data, self.schema, index, self.tree)

    def leaves(self, data, schema, tree):
        for item_index in range(self.index, len(data)):
            self.enforce(data, schema, item_index, tree)

    def enforce(self, data, schema, item_index, tree):
        try:
            _validate = Validator({}, self.schema)
            _validate.data = {0: data[item_index]}
            _validate.validate()
        except Invalid:
            e = sys.exc_info()[1]
            tree.extend(e.path)
            raise Invalid(e.schema_item, tree, pair='value')
        except SchemaError:
            e = sys.exc_info()[1]
            tree.extend(e.path)
            raise SchemaError('', tree, reason=e._reason, pair='value')


def normalize(data_structure, sort=True):
    """
    Receives a dictionary and returns an ordered
    dictionary, with integers as keys for tuples.
    """
    if sort:
        for k, v in data_structure.items():
            if isinstance(v, dict):
                data_structure[k] = normalize(v)
        data_structure = sorted(data_structure.items())
    return dict((number, value) for number, value in enumerate(data_structure))


def normalize_schema(data_structure):
    if len(data_structure) == 2 and isinstance(data_structure[1], tuple) or len(data_structure) > 2:
        if not isinstance(data_structure[0], tuple):
            new_struct = {0: (data_structure[0], normalize_schema(data_structure[1]))}
        else:
            new_struct = dict((number, value) for number, value in enumerate(data_structure))
        for i in range(len(new_struct)):
            value = new_struct.get(i)
            if len(value) == 2 and isinstance(value[1], tuple): # a nested tuple
                new_struct[i] = (value[0], normalize_schema(value[1]))
        return new_struct
    else:
        return {0: data_structure}


def enforce(data_item, schema_item, tree, pair):
    if is_callable(schema_item):
        try:
            schema_item(data_item)
        except AssertionError:
            e = sys.exc_info()[1]
            raise Invalid(schema_item, tree, reason=e, pair=pair)
    else:
        try:
            assert data_item == schema_item
        except AssertionError:
            if pair == 'value':
                tree.append(data_item)
            raise Invalid(schema_item, tree, pair=pair)


def validate(data, schema):
    """
    Main entry point for the validation engine.

    :param data: The incoming data, as a dictionary object.
    :param schema: The schema from which data will be validated against
    """
    validator = Validator(data, schema)
    validator.validate()



from notario.exceptions import Invalid
from notario.engine import RecursiveValidator


class BasicRecursiveValidator(object):
    """
    Base class for recursive validators, can be sub-classed
    for other type of recursive validators but should not be
    used directly.
    """

    __validator_leaf__ = True

    def __init__(self, schema):
        self.schema = schema


class AnyObject(BasicRecursiveValidator):
    """
    Go over all the items in an dict object and make sure that at least
    one of the items validates correctly against the schema provided.
    If no items pass it raises ``Invalid``.


    .. testsetup:: anyobject

        from notario import validate
        from notario.validators.recursive import AnyObject

    Example usage for single values::

        data = {'foo': {{'a':10}, {'b':20}, {'c':40}}}
        schema = ('foo', AnyObject(('a', 10)))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'b': 10}, {'a': 1}]}
        schema = ('foo', AnyObject(('a', 1))
        validate(data, schema)

    When a single item in the array fails to pass against the validator's schema it
    stops further iteration and it will raise an error like:


    .. doctest:: anyobject

        >>> data = {'foo': {'a':{'a':10}, 'b':{'a':20}, 'c':{'a':20}}}
        >>> schema = ('foo', AnyObject(('a', ('a', 90))))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: foo -> a  did not contain any valid objects against callable: AnyObject

    In this particular validator, it remembers on what key of the dict object
    the failure was created and it goes even further giving the key and value
    of the object it went against.
    """

    def __call__(self, data, tree):
        index = len(data) - 1
        validator = RecursiveValidator(data, self.schema, [], index=index)
        for item_index in range(len(data)):
            try:
                return validator.leaf(item_index)
            except Invalid:
                if tree:
                    tree.pop
                pass

        msg = "did not contain any valid objects against callable: %s" % self.__class__.__name__
        raise Invalid(self.schema, tree, pair='value', msg=msg)



class AllObjects(BasicRecursiveValidator):
    """
    For all the objects contained in a dictionary apply the schema passed in
    to the validator.
    If a single item object fails, it raises ``Invalid``.


    .. testsetup:: allobjects

        from notario import validate
        from notario.validators.recursive import AllObjects

    Example usage for single values::

        data = {'foo': {{'a':10}, {'a':20}, {'a':20}}}
        schema = ('foo', AllObjects(('a', 20)))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'a': 1}, {'a': 1}]}
        schema = ('foo', AllObjects(('a', 1))
        validate(data, schema)

    When a single item in the array fails to pass against the validator's schema it
    stops further iteration and it will raise an error like:


    .. doctest:: allobjects

        >>> data = {'foo': {'a':{'a':10}, 'b':{'a':20}, 'c':{'a':20}}}
        >>> schema = ('foo', AllObjects(('a', ('a', 90))))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> a -> a -> 10  did not match 90

    In this particular validator, it remembers on what key of the dict object
    the failure was created and it goes even further giving the key and value
    of the object it went against.


    """

    def __call__(self, data, tree):
        validator = RecursiveValidator(data, self.schema, tree)
        validator.validate()





def is_callable(data):
    if hasattr(data, '__call__'):
        return True
    return False


def optional(validator):
    """
    This decorator allows to have validators work only when there is
    a value that contains some data, otherwise it will just not pass
    the information to the actual validator and will not fail as a
    result.

    As any normal decorator, it can be used corectly with the decorator
    syntax or in the actual schema.

    This is how it would look in a schema::

        ('key', optional(my_validator))

    Where ``my_validator`` can be any validator that accepts a single
    argument.

    In case a class based validator is being used (like the ``recursive`` or ``iterables``
    then it would look like::

        ('key', optional(class_validator(('key', 'value'))))

    Of course, the schema should vary depending on your needs, it is just the
    way of constructing the validator call that should be important.
    """

    def decorated(value):
        if value:
            return validator(value)
        return

    return decorated



from notario import exceptions


def foo(): return True

class Object(object): pass

class TestInvalid(object):

    def test_include_the_key(self):
        error = exceptions.Invalid('key', ['foo', 'bar', 'key'])
        assert 'key' in error._format_path()

    def test_include_the_path_in_str(self):
        error = exceptions.Invalid('key', ['path'])
        assert 'path' in error.__str__()

    def test_include_the_key_in_str(self):
        error = exceptions.Invalid('key', ['path'])
        assert 'key' in error.__str__()

    def test_multiple_keys_in_format_path(self):
        error = exceptions.Invalid('schema', ['key', 'subkey', 'bar'])
        assert '-> key -> subkey -> bar' in error._format_path()

    def test_full_message(self):
        error = exceptions.Invalid('3', ['foo', 'bar', 'baz'])
        result = error.__str__()
        assert "-> foo -> bar -> baz key did not match '3'" == result

    def test_full_message_for_callable(self):
        error = exceptions.Invalid(foo, ['foo', 'bar', 'baz'])
        result = error.__str__()
        assert "-> foo -> bar -> baz key did not pass validation against callable: foo" == result

    def test_full_message_for_value(self):
        error = exceptions.Invalid('3', ['foo', 'bar', 'baz'], pair='value')
        result = error.__str__()
        assert "-> foo -> bar -> baz  did not match '3'" == result

    def test_full_message_for_callable_with_value(self):
        error = exceptions.Invalid(foo, ['foo', 'bar', 'baz'], pair='value')
        result = error.__str__()
        assert "-> foo -> bar -> baz  did not pass validation against callable: foo" == result



from pytest import raises
from notario.validators import types
from notario.validators import recursive
from notario.exceptions import Invalid


class TestAllObjects(object):

    def test_all_objects_fail(self):
        data = {0:('a', '1'), 1:('b', 2), 2:('c', '3')}
        schema = ((types.string, types.integer))
        with raises(Invalid) as exc:
            any_object = recursive.AllObjects(schema)
            any_object(data, [])
        msg = '-> a  did not pass validation against callable: integer'
        assert exc.value.args[0] == msg

    def test_all_objects_pass(self):
        data = {0:('a', 1), 1:('b', 2), 2:('c', 3)}
        schema = ((types.string, types.integer))
        any_object = recursive.AllObjects(schema)
        assert any_object(data, []) is None


class TestAnyObject(object):

    def test_any_object_fail_all(self):
        data = {0:('a', '1'), 1:('b', '2'), 2:('c', '3')}
        schema = ((types.string, types.integer))
        with raises(Invalid) as exc:
            any_object = recursive.AnyObject(schema)
            any_object(data, [])
        msg = '-> top level did not contain any valid objects against callable: AnyObject'
        assert exc.value.args[0] == msg

    def test_any_object_pass_first(self):
        data = {0:('a', '1'), 1:('b', 2), 2:('c', 3)}
        schema = ((types.string, types.string))
        any_object = recursive.AnyObject(schema)
        assert any_object(data, []) is None

    def test_any_object_pass_second(self):
        data = {0:('a', 1), 1:('b', '2'), 2:('c', 3)}
        schema = ((types.string, types.string))
        any_object = recursive.AnyObject(schema)
        assert any_object(data, []) is None

    def test_any_object_pass_last(self):
        data = {0:('a', 1), 1:('b', 2), 2:('c', '3')}
        schema = ((types.string, types.string))
        any_object = recursive.AnyObject(schema)
        assert any_object(data, []) is None



from pytest import raises
from notario import engine
from notario.exceptions import Invalid, SchemaError
from notario.validators import recursive, iterables, types


class TestEnforce(object):

    def test_callable_failure_raises_invalid(self):
        def bad_callable(v): assert False
        with raises(Invalid) as exc:
            engine.enforce('1', bad_callable, ['1'], 'key')
        assert exc.value.args[0] == '-> 1 key did not pass validation against callable: bad_callable'

    def test_callable_passes_with_flying_colors(self):
        def cool_callable(v): pass
        result = engine.enforce('1', cool_callable, ['1'], 'key')
        assert result is None

    def test_equal_params_pass_with_flying_colors(self):
        result = engine.enforce('1', '1', ['1'], 'key')
        assert result is None

    def test_unequal_params_raise_invalid(self):
        with raises(Invalid) as exc:
            engine.enforce(1, 2, ['1'], 'key')
        assert exc.value.args[0] == '-> 1 key did not match 2'

    def test_callable_with_messages_are_passed_on(self):
        def callable_message(v): assert False, "this is completely False"
        with raises(Invalid) as exc:
            engine.enforce(1, callable_message, ['1'], 'key')
        assert exc.value.reason == "this is completely False"


class TestNormalizeSchema(object):

    def test_return_data_structure_with_cero_index(self):
        result = engine.normalize_schema(['a'])
        assert result == {0: ['a']}

    def test_recurse_if_second_item_is_typle(self):
        data = ('a', ('a', 'b'))
        result = engine.normalize_schema(data)
        assert result == {0: ('a', {0: ('a', 'b')})}

    def test_respect_more_than_two_values_in_tuple(self):
        data = ('a', (('a', 'b'), ('c', 'c'), ('d', 'd')))
        result = engine.normalize_schema(data)
        assert result == {0: ('a', {0: ('a', 'b'), 1: ('c', 'c'), 2: ('d', 'd')})}

    def test_basic_key_value_pairs(self):
        data = (('a', 'b'), ('b', 'b'), ('c', 'c'))
        result = engine.normalize_schema(data)
        assert result == {0: ('a', 'b'), 1: ('b', 'b'), 2: ('c', 'c')}

class TestValidator(object):

    def test_validate_top_level_keys(self):
        data = {'a': 1, 'b': 2}
        schema = (('a', 1), ('c', 2))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == "-> b key did not match 'c'"

    def test_validate_length_equality(self):
        data = {'a': 1, 'b': 2}
        schema = (('a', 1, 2), ('c', 2))
        with raises(SchemaError) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> top level length did not match schema'

    def test_validate_length_equality_less_items(self):
        data = {'a': 1, 'b': 2, 'c':'c'}
        schema = (('a', 1), ('b', 2))
        with raises(SchemaError) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> top level has less items in schema than in data'

    def test_validate_length_equality_returns(self):
        data = {0:('a', 1)}
        class _Schema(object):
            __validator_leaf__ = True
        schema = {0:_Schema()}
        validator = engine.Validator({}, {})
        result = validator.length_equality(data, schema, 0, [])
        assert result is None

    def test_validate_top_level_values(self):
        data = {'a': 1, 'b': 2}
        schema = (('a', 1), ('b', 1))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> b -> 2  did not match 1'

    def test_validate_value_nested_dictionaries(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : 1}}
        schema = (('a', 1), ('b', (('a', 2), ('b', 2))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> b -> b -> 1  did not match 2'

    def test_validate_key_doubly_nested_dictionaries(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : {'a':'a'}}}
        schema = (('a', 1), ('b', (('a', 2), ('b', ('a', 'b')))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == "-> b -> b -> a -> a  did not match 'b'"

    def test_validate_key_nested_dictionaries(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : 1}}
        schema = (('a', 1), ('b', (('b', 2), ('b', 1))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == "-> b -> a key did not match 'b'"

    def test_validate_value_nested_arrays(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : [1, 2, 3]}}
        schema = (('a', 1), ('b', (('a', 2), ('b', [1, 1, 3]))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> b -> b -> [1, 2, 3]  did not match [1, 1, 3]'

    def test_validate_key_nested_arrays(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : [1, 2, 3]}}
        schema = (('a', 1), ('b', (('a', 2), ('b', [1, 1, 3]))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> b -> b -> [1, 2, 3]  did not match [1, 1, 3]'

    def test_validate_multiple_items_as_values(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : 1, 'd':1, 'c':2}}
        schema = (('a', 1), ('b', (('a', 2), ('b', 1), ('c', 2), ('d', 2))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> b -> d -> 1  did not match 2'


class TestValidatorLeaves(object):
    """
    The Validator object would behave differently with
    validator classes that have  the __validator_leaf__
    """

    def test_traverser_returns_the_iterable_leaf_if_seen(self):
        data = {'a': {'b': [1, 1, 1, 1]}}
        schema = ('a', ('b', iterables.AllItems(2)))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> a -> b -> list[0] item did not match 2'

    def test_traverser_returns_the_recursive_leaf_if_seen(self):
        data = {'a': {'b': [1, 1, 1, 1]}}
        schema = ('a', recursive.AllObjects(('b', [1, 1, 1, 2])))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> a -> b -> [1, 1, 1, 1]  did not match [1, 1, 1, 2]'


class TestRecursiveValidator(object):

    def test_bad_index_number(self):
        data = {'a': {'a':'a', 'b':'b'}}
        schema = ('a', 'a')
        recursive_validator = recursive.RecursiveValidator(data, schema, index=100)
        with raises(SchemaError) as exc:
            recursive_validator.validate()

        assert exc.value.args[0] == '-> top level has not enough items to select from'

    def test_deal_with_recursion(self):
        data = {'a': {'a':'a', 'b':{'a': 'b', 'c':'c', 'd':1}}}
        schema = ('a', (('a', 'a'), ('b', recursive.AllObjects((types.string, types.string)))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        assert exc.value.args[0] == '-> a -> b -> d  did not pass validation against callable: string'


class TestIterableValidator(object):

    def test_not_enough_items_in_data(self):
        data = {0: ('a', 'b')}
        schema = {0: ('a', 'b')}
        iter_validator = engine.IterableValidator(data, schema, index=100)
        with raises(SchemaError) as exc:
            iter_validator.validate()

        assert exc.value.args[0] == '-> top level has not enough items to select from'

    def test_validate_nested_array_first(self):
        data = [{'a':'a'}, {'b': 'b'}]
        schema = (types.string, 'b') # note how this is not normalized
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(Invalid) as exc:
            iter_validator.validate()

        assert exc.value.args[0] == "-> list[0] -> a -> a  did not match 'b'"

    def test_validate_nested_array(self):
        data = [{'a':'b'}, {'b': 'c'}]
        schema = (types.string, 'b') # note how this is not normalized
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(Invalid) as exc:
            iter_validator.validate()

        assert exc.value.args[0] == "-> list[1] -> b -> c  did not match 'b'"

    def test_report_schema_errors(self):
        data = [1,2,3,4]
        schema = (types.string, ('b', 'a'))
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(SchemaError) as exc:
            iter_validator.validate()

        assert exc.value.args[0] == "-> top level iterable contains single items, schema does not"

    def test_schema_error_on_non_list(self):
        data = {'foo': [{'a':'b'}, {'b': 'c'}]}
        schema = (types.string, 'b') # note how this is not normalized
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(SchemaError) as exc:
            iter_validator.validate()

        assert exc.value.args[0] == "-> top level IterableValidator needs a list to validate"



from notario.utils import is_callable


class NotarioException(Exception):

    def __init__(self, schema_item, path, reason=None, pair='key', msg=None):
        self.schema_item = schema_item
        self.path = path
        self._reason = reason
        self._pair = pair
        self._msg = msg
        Exception.__init__(self, self.__str__())

    def __str__(self):
        return self._get_message()

    def _format_path(self, use_pair=True):
        message = ""
        for key in self.path:
            accessed_key = '-> %s ' % key
            message += accessed_key
        if use_pair and self._pair != 'value':
            return message + self._pair
        return message or "-> top level"

    def _format_message(self):
        if self._msg:
            return self._msg
        if is_callable(self.schema_item):
            msg = "did not pass validation against callable: %s" % (self.schema_item.__name__)
        else:
            msg = "did not match %s" % (repr(self.schema_item))
        return msg

    @property
    def reason(self):
        try:
            return self._reason.args[0]
        except IndexError:
            return self._reason



class Invalid(NotarioException):
    """
    This Exception class is used only by the :class:`Validator`
    class to provide a clear message when a validation mismatch occurs
    while traversing the configuration tree.
    """

    def _get_message(self):
        return "%s %s" % (self._format_path(), self._format_message())



class SchemaError(NotarioException):

    def _get_message(self):
        msg = "%s %s" % (self._format_path(use_pair=False), self._reason)
        return msg



"""
Iterable validators for array objects only. They provide a way of
applying a schema to any given items in an array.
"""
from notario.exceptions import Invalid
from notario.engine import IterableValidator
from notario.utils import is_callable


class BasicIterableValidator(object):
    """
    Base class for iterable validators, can be sub-classed
    for other type of iterable validators but should not be
    used directly.
    """

    __validator_leaf__ = True

    def __init__(self, schema):
        self.schema = schema


class AnyItem(BasicIterableValidator):
    """
    Go over all the items in an array and make sure that at least
    one of the items validates correctly against the schema provided.
    If no items pass it raises ``Invalid``.

    .. note:: It only works on arrays, otherwise it will raise a ``SchemaError``

    .. testsetup::

        from notario import validate
        from notario.validators.iterables import AnyItem

    Example usage for single values::

        data = {'foo' : [10, 30, 50]}
        schema = ('foo', AnyItem(50))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'a': 1}, {'b': 2}]}
        schema = ('foo', AnyItem(('b', 2))
        validate(data, schema)

    When a single item in the array matches correctly against the validator's schema it
    stops further iteration and the validation passes. Otherwise it will raise an error
    like:


    .. doctest::

        >>> data = {'foo': [{'a': 1}, {'b': 2}]}
        >>> schema = ('foo', AnyItem(('c', 4)))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> list[]  did not contain any valid items matching ('c', 4)

    """

    def __call__(self, data, tree):
        index = len(data) - 1
        validator = IterableValidator(data, self.schema, [], index=index)
        for item_index in range(len(data)):
            try:
                return validator.leaf(item_index)
            except Invalid:
                pass

        tree.append('list[]')
        if is_callable(self.schema):
            msg = "did not contain any valid items against callable: %s" % self.schema.__name__
        else:
            msg = "did not contain any valid items matching %s" % repr(self.schema)
        raise Invalid(self.schema, tree, pair='value', msg=msg)


class AllItems(BasicIterableValidator):
    """
    For all the items in an array apply the schema passed in to the validator.
    If a single item fails, it raises ``Invalid``.

    .. note:: It only works on arrays, otherwise it will raise a ``SchemaError``

    .. testsetup:: allitems

        from notario import validate
        from notario.validators.iterables import AllItems

    Example usage for single values::

        data = {'foo' : [10, 10, 10]}
        schema = ('foo', AllItems(10))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'a': 1}, {'a': 1}]}
        schema = ('foo', AllItems(('a', 1))
        validate(data, schema)

    When a single item in the array fails to pass against the validator's schema it
    stops further iteration and it will raise an error like:


    .. doctest:: allitems

        >>> data = {'foo': [{'a': 1}, {'a': 2}]}
        >>> schema = ('foo', AllItems(('a', 1)))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> list[1] -> a -> 2  did not match 1

    In this particular validator, it remembers on what index of the array the failure
    was created and it goes even further giving the key and value of the object it
    went against.

    """

    def __call__(self, data, tree):
        validator = IterableValidator(data, self.schema, tree)
        validator.validate()



"""
Basic type validators
"""
from notario._compat import basestring

def string(value):
    """
    Validates a given input is of type string.

    Example usage::

        data = {'a' : 21}
        schema = (string, 21)
    """
    assert isinstance(value, basestring), "not of type string"


def boolean(value):
    """
    Validates a given input is of type boolean.

    Example usage::

        data = {'a' : True}
        schema = ('a', boolean)
    """
    assert isinstance(value, bool), "not of type boolean"


def dictionary(value):
    """
    Validates a given input is of type dictionary.

    Example usage::

        data = {'a' : {'b': 1}}
        schema = ('a', dictionary)
    """
    assert isinstance(value, dict), "not of type dictionary"


def array(value):
    """
    Validates a given input is of type list.

    Example usage::

        data = {'a' : [1,2]}
        schema = ('a', array)
    """
    assert isinstance(value, list), "not of type array"

def integer(value):
    """
    Validates a given input is of type int..

    Example usage::

        data = {'a' : 21}
        schema = ('a', integer)
    """
    assert isinstance(value, int), "not of type int"



from notario import utils


class TestIsCallable(object):

    def setup(self):
        def fake_callable(): pass
        self.fake_callable = fake_callable

    def test_is_callable(self):
        result = utils.is_callable(self.fake_callable)
        assert result is True

    def test_is_not_callable(self):
        result = utils.is_callable(1)
        assert result is False


class TestOptionalDecorator(object):

    def fake_validator(self, value):
        return True

    def test_has_no_value(self):
        result = utils.optional(self.fake_validator)('')
        assert result is None

    def test_value(self):
        result = utils.optional(self.fake_validator)('some value!')
        assert result is True



from pytest import raises
from notario.validators import types

#
# Most of these are just excercising the code really
# as they are just making sure isinstance returns
# what we expect, effectively testing python
#

class TestTypes(object):

    def test_string_pass(self):
        assert types.string('a string') is None

    def test_boolean_pass(self):
        assert types.boolean(False) is None

    def test_dictionary_pass(self):
        assert types.dictionary({}) is None 

    def test_array_pass(self):
        assert types.array([]) is None 

    def test_integer_pass(self):
        assert types.integer(1) is None 

    def test_string_fail(self):
        with raises(AssertionError) as exc:
            types.string(1) 
        assert exc.value.args[0] == 'not of type string'

    def test_boolean_fail(self):
        with raises(AssertionError) as exc:
            types.boolean('a string') 
        assert exc.value.args[0] == 'not of type boolean'

    def test_dictionary_fail(self):
        with raises(AssertionError) as exc:
            types.dictionary([])  
        assert exc.value.args[0] == 'not of type dictionary'

    def test_array_fail(self):
        with raises(AssertionError) as exc:
            types.array({})  
        assert exc.value.args[0] == 'not of type array'

    def test_integer_fail(self):
        with raises(AssertionError) as exc:
            types.integer('a string')  
        assert exc.value.args[0] == 'not of type int'



from pytest import raises
from notario.validators import iterables
from notario.exceptions import Invalid, SchemaError


class TestAnyItem(object):

    def test_any_item_fail(self):
        data = [1, 2, 6, 4, 5]
        schema = 7
        with raises(Invalid) as exc:
            any_item = iterables.AnyItem(schema)
            any_item(data, [])
        msg = '-> list[]  did not contain any valid items matching 7'
        assert exc.value.args[0] == msg

    def test_tuple_schemas_cannot_run_on_single_values(self):
        data = [1, 2, 6, 4, 5]
        schema = ('a', 'b')
        with raises(SchemaError) as exc:
            any_item = iterables.AnyItem(schema)
            any_item(data, [])
        msg = "-> top level iterable contains single items, schema does not"
        assert exc.value.args[0] == msg

    def test_single_schemas_can_run_on_tuple_items(self):
        data = [{'a':'a'}, {'b':'b'}]
        schema = 'a'
        with raises(Invalid) as exc:
            any_item = iterables.AnyItem(schema)
            any_item(data, [])
        msg = "-> list[]  did not contain any valid items matching 'a'"
        assert exc.value.args[0] == msg


    def test_any_item_pass(self):
        data = [1, 2, 6, 4, 5]
        schema = 6
        any_item = iterables.AnyItem(schema)
        assert  any_item(data, []) is None

    def test_any_item_pass_last_item(self):
        data = [1, 2, 3, 4, 6]
        schema = 6
        any_item = iterables.AnyItem(schema)
        assert  any_item(data, []) is None


class TestAllItems(object):

    def test_all_items_pass(self):
        data = [1, 1, 1, 1]
        schema = 1
        all_items = iterables.AllItems(schema)
        assert all_items(data, []) is None

    def test_all_items_fail(self):
        data = [1, 1, 3, 1]
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        assert exc.value.args[0] == '-> list[2] item did not match 1'

    def test_all_items_fail_last_item(self):
        data = [1, 1, 1, 3]
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        assert exc.value.args[0] == '-> list[3] item did not match 1'

    def test_all_items_fail_uses_first_invalid(self):
        data = [3, 5, 2, 3]
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        assert exc.value.args[0] == '-> list[0] item did not match 1'




from pytest import raises
from notario import validate
from notario.exceptions import Invalid, SchemaError
from notario.validators import iterables, recursive, types


class TestValidate(object):

    def test_most_simple_validation(self):
        data = {'a': 'a'}
        schema = (('a', 'b'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> a -> a  did not match 'b'"
        
    def test_multi_pair_non_nested_last(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'c'), ('d', 'a'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> d -> d  did not match 'a'"

    def test_multi_pair_non_nested_first(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'b'), ('b', 'b'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> a -> a  did not match 'b'"

    def test_multi_pair_non_nested_second(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'a'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> b -> b  did not match 'a'"

    def test_multi_pair_non_nested_third(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'a'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> c -> c  did not match 'a'"

    def test_multi_pair_non_nested_last_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'c'), ('a', 'a'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> d key did not match 'a'"

    def test_multi_pair_non_nested_first_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('f', 'a'), ('b', 'b'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> a key did not match 'f'"

    def test_multi_pair_non_nested_second_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('f', 'b'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> b key did not match 'f'"

    def test_multi_pair_non_nested_third_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('f', 'c'), ('d', 'a'))

        with raises(Invalid) as exc:
            validate(data, schema)

        assert exc.value.args[0] == "-> c key did not match 'f'"


class TestWithIterableValidators(object):

    def test_all_items_pass(self):
        data = {'a': [1, 2, 3, 4, 5]}
        schema = ('a', iterables.AllItems(types.integer))
        assert validate(data, schema) is None

    def test_any_items_pass(self):
        data = {'a': [1,2, 'a string' ,4,5]}
        schema = ('a', iterables.AnyItem(types.string))
        assert validate(data, schema) is None

    def test_all_items_fail(self):
        data = {'a': [1, 2, '3', 4, 5]}
        schema = ('a', iterables.AllItems(types.integer))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a -> list[2] item did not pass validation against callable: integer'

    def test_all_items_fail_length(self):
        data = {'a': [{'a': 2}, {'b': {'a': 'b'}}]}
        schema = ('a', iterables.AllItems((types.string, 2)))
        with raises(SchemaError) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a -> b  has less items in schema than in data'

    def test_all_items_fail_non_callable(self):
        data = {'a': [1, 2, '3', 4, 5]}
        schema = ('a', iterables.AllItems('foo'))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == "-> a -> list[0] item did not match 'foo'"


    def test_any_items_fail(self):
        data = {'a': [1, 2, 3, 4, 5]}
        schema = ('a', iterables.AnyItem(types.string))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a -> list[]  did not contain any valid items against callable: string'

    def test_any_items_fail_non_callable(self):
        data = {'a': [1, 2, 3, 4, 5]}
        schema = ('a', iterables.AnyItem('foo'))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == "-> a -> list[]  did not contain any valid items matching 'foo'"

    def test_any_item_with_dictionaries(self):
        data = {'a': [{'a': 1}, {'b': 2}]}
        schema = ('a', iterables.AnyItem(('c', 4)))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == "-> a -> list[]  did not contain any valid items matching ('c', 4)"


class TestWithRecursiveValidators(object):

    def test_all_objects_pass(self):
        data = {'a': {'a': 1, 'b': 2, 'c': 3}}
        schema = ('a', recursive.AllObjects((types.string, types.integer)))
        assert validate(data, schema) is None

    def test_any_objects_pass(self):
        data = {'a': {'a': 1, 'b':'a string', 'c': 3}}
        schema = ('a', recursive.AnyObject((types.string, types.string)))
        assert validate(data, schema) is None

    def test_all_objects_fail(self):
        data = {'a': {'a': 1, 'b': 'a string', 'c': 3}}
        schema = ('a', recursive.AllObjects((types.string, types.integer)))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a -> b  did not pass validation against callable: integer'

    def test_all_objects_fail_non_callable(self):
        data = {'a': {'a': 1, 'b': 1, 'c': 1}}
        schema = ('a', recursive.AllObjects((types.string, 2)))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a -> a -> 1  did not match 2'

    def test_all_objects_fail_length(self):
        data = {'a': {'a': 2, 'b': {'a': 'b'}}}
        schema = ('a', recursive.AllObjects((types.string, 2)))
        with raises(SchemaError) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a -> b  has less items in schema than in data'

    def test_any_object_fail(self):
        data = {'a': {'a': 1, 'b': 4, 'c': 3}}
        schema = ('a', recursive.AnyObject((types.string, types.string)))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a  did not contain any valid objects against callable: AnyObject'

    def test_any_objects_fail_non_callable(self):
        data = {'a': {'a': 1, 'b': 4, 'c': 3}}
        schema = ('a', recursive.AnyObject(('a', 'a')))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> a  did not contain any valid objects against callable: AnyObject'

    def test_no_pollution_from_previous_traversing_all_objects(self):
        data = {
                'a' : { 'a': 1, 'b': 2 },
                'b' : { 'c' :1, 'd': 2 },
                'c' : { 'e': 1, 'f': 2 }
            }
        schema = (
                ('a', (('a', 1), ('b', 2))),
                ('b', (('c', 1), ('d', 2))),
                ('c', recursive.AllObjects((types.string, types.string)))
            )
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> c -> e  did not pass validation against callable: string'

    def test_no_pollution_from_previous_traversing_all_items(self):
        data = {
                'a' : { 'a': 1, 'b': 2 },
                'b' : { 'c' :1, 'd': 2 },
                'c' : ['e', 1, 'f', 2 ]
            }
        schema = (
                ('a', (('a', 1), ('b', 2))),
                ('b', (('c', 1), ('d', 2))),
                ('c', iterables.AllItems(types.string))
            )
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.args[0] == '-> c -> list[1] item did not pass validation against callable: string'




import re


module_file = open("notario/__init__.py").read()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))
long_description = open('README.rst').read()

from setuptools import setup, find_packages

setup(
    name             = 'notario',
    description      = 'A dictionary validator',
    packages         = find_packages(),
    author           = 'Alfredo Deza',
    author_email     = 'alfredodeza [at] gmail.com',
    version          = metadata['version'],
    url              = 'http://github.com/alfredodeza/notario',
    license          = "MIT",
    zip_safe         = False,
    keywords         = "schema, validator, dictionary, enforce",
    long_description = long_description,
    classifiers      = [
                        'Development Status :: 4 - Beta',
                        'Intended Audience :: Developers',
                        'License :: OSI Approved :: MIT License',
                        'Topic :: Software Development :: Build Tools',
                        'Topic :: Software Development :: Libraries',
                        'Topic :: Software Development :: Testing',
                        'Topic :: Utilities',
                        'Operating System :: MacOS :: MacOS X',
                        'Operating System :: Microsoft :: Windows',
                        'Operating System :: POSIX',
                        'Programming Language :: Python :: 2.6',
                        'Programming Language :: Python :: 2.7',
                        'Programming Language :: Python :: 3.0',
                        'Programming Language :: Python :: 3.1',
                        'Programming Language :: Python :: 3.2',
                        'Programming Language :: Python :: Implementation :: PyPy'
                      ]
)
